﻿//Release Info
//V1.2
//  Updated to handle labels suggested by Timbo.
//  Removed "Reset Trackers" button redundant, should automatically detect trackers. If you need the feature the similar action is reload plugin.
//
//v1.1 9/5/2020
//  Initial release.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;

namespace VRStudyPlugin
{
    class LabelTrackers : MVRScript
    {
        #region PluginInfo    
        public string pluginAuthor = "VRStudy";
        public string pluginName = "LabelTrackers";
        public string pluginVersion = "1.2";
        public string pluginDate = "9/11/2020";
        public string pluginDescription =
 @"     A plugin to make it easier to find the name of trackers in VR. Attach this as a scene plugin.
        NOTE: This will only work in VR with trackers present.";
        #endregion

        #region TrackerData
              
        public class TrackerData
        {
            private GameObject _gameObj;
            private TextMesh _tm;
            private Transform _tracker;
            private bool _showRole;

            public bool hide;
            public bool showRole;
            public bool showLabel;
            public string role;
            public float FontSize { get { return _tm.fontSize; } set { SetFontSize(value); } }
            public JSONStorableString trackerLabel;

            public TrackerData(Transform tracker,string role="")
            {
                this._tracker = tracker;
                this._showRole = true;
                this.showLabel = false;
                this.role = role;
                
                SetupTracker();
            }

            public void SetupTracker()
            {                               
                //Add a game object for tracker
                _gameObj = new GameObject();
                _gameObj.name = $"td_{_tracker.name}";

                //Add the tracker label to it
                _tm = _gameObj.AddComponent<TextMesh>();
                _tm.alignment = TextAlignment.Center;                                
                _tm.characterSize = 0.01f;
                _tm.fontSize = 50;
                _tm.richText = true;

                //Set the trackers Roles/Label
                trackerLabel = new JSONStorableString(_tracker.name, role);
                trackerLabel.altName = role;

                SetTrackerText();
                DoUpdate();
            }
            
            //Displays tracker text
            public void SetTrackerText()
            {
                string displayText = "";
                if (!showLabel)
                {
                    displayText = $"{trackerLabel.name}";
                    if (showRole) displayText += $"\n<size=25>[{trackerLabel.val}]</size>";
                }
                else
                {
                    displayText = $"{trackerLabel.val}";
                }
                _tm.text = displayText;
            }

            //Sets the size of label
            private void SetFontSize(float fSize = 50f)
            {
                if(_tm!=null) _tm.fontSize = Mathf.RoundToInt(fSize);
            }
            
            //Set the tracker text to be hidden
            public void SetActive(bool active = true)
            {
                _gameObj?.SetActive(active);
            }

            //Update gameobject to follow active tracker
            public void DoUpdate()
            {
                //Set only active trackers to be visible                
                if (!hide)
                {
                    //This will handle if a tracker is turned on/off during play
                    SetActive(_tracker.gameObject.activeSelf);
                }
                else
                {
                    SetActive(false);
                }
                
                //Only track if the tracker is active
                if (_tracker.gameObject.activeSelf)
                {
                    _gameObj.transform.position = _tracker.position;
                    _gameObj.transform.localRotation = SuperController.singleton.lookCamera.transform.rotation;
                }
            }

            public void DoDestroy()
            {
                //Make sure we delete the game object or Unity/VAM will most likely 
                //crash when reloading script
                if (_gameObj != null) Destroy(_gameObj);
            }
        }
        #endregion TrackerData
        
        #region TrackTrackers
        public class TrackTrackers
        {
            //Hold a list with information of each tracker
            private List<TrackerData> trackers = new List<TrackerData>();
            
            //Tracker roles match VAM's settings
            private readonly Dictionary<int, string> _trackerRoles = new Dictionary<int, string>
            {
                [1] = "LeftFoot",
                [2] = "RightFoot",
                [3] = "LeftShoulder",
                [4] = "RightShoulder",
                [5] = "Waist",
                [6] = "Chest",
                [7] = "Camera",
                [8] = "Keyboard"
            };

            public TrackTrackers()
            {
                GetTrackers();
            }

            //Find each tracker
            public void GetTrackers()
            {
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker1,_trackerRoles[1]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker2,_trackerRoles[2]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker3,_trackerRoles[3]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker4,_trackerRoles[4]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker5,_trackerRoles[5]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker6,_trackerRoles[6]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker7,_trackerRoles[7]));
                trackers.Add(new TrackerData(SuperController.singleton.viveTracker8,_trackerRoles[8]));
            }

            //Gets a list of all trackers as storable string
            public List<JSONStorableString> ToStorableString()
            {
                List<JSONStorableString> tempList = new List<JSONStorableString>();                
                foreach (TrackerData td in trackers)
                {
                    tempList.Add(td.trackerLabel);//ToStorableString());
                }
                return tempList;
            }

            //Set tracker visibility
            public void SetHidden(bool hide = false)
            {
                SuperController.singleton.hideTrackers = hide;
                foreach (TrackerData td in trackers)
                {
                    td.hide=hide;
                }
            }
            
            //Used for reflecting changes to text input
            public void UpdateText()
            {
                foreach(TrackerData td in trackers)
                {
                    td.SetTrackerText();
                }
            }

            //Show or hide the role under the name
            public void ShowRole(bool showRole=true)
            {
                foreach(TrackerData td in trackers)
                {
                    td.showRole=showRole;
                    td.SetTrackerText();
                }
            }
            
            //Show label/role only
            public void ShowLabel(bool showLabel = false)
            {
                foreach (TrackerData td in trackers)
                {
                    td.showLabel = showLabel;
                    td.SetTrackerText();
                }
            }

            //Sets the font size for each tracker
            public void SetFontSize(float fSize)
            {
                foreach(TrackerData td in trackers)
                {
                    td.FontSize=fSize;
                }
            }

            //Update the position and rotation of text
            public void DoUpdate()
            {
                foreach(TrackerData td in trackers)
                {
                    td.DoUpdate();
                }
            }
            
            //Clean up old data
            public void DoDestroy()
            {
                foreach (TrackerData td in trackers)
                {
                    td.DoDestroy();
                }
                trackers.Clear();
            }
        }
        #endregion TrackTrackers

        #region SetupUI
                
        protected JSONStorableBool bHideTrackers;
        protected JSONStorableBool bHideText;
        protected JSONStorableFloat fontSize;
        protected JSONStorableBool bShowRole;
        protected JSONStorableBool bShowLabel;
        protected UIDynamicToggle toggleShowRole;
        protected UIDynamicToggle toggleShowLabel;
        protected List<UIDynamic> trackerLabels = new List<UIDynamic>();

        public void SetupUI()
        {
            //Toggle the trackers on/off
            bHideTrackers = new JSONStorableBool("Hide trackers", false,
                (bool bHide) =>
                {
                    trackEm.SetHidden(bHide);
                });
            RegisterBool(bHideTrackers);
            CreateToggle(bHideTrackers);

            //Sets the font size for the tracker labels
            fontSize = new JSONStorableFloat("Font Size", 50,
                (float fSize) =>
                {
                    trackEm.SetFontSize(fSize);
                },
                1, 200);
            RegisterFloat(fontSize);
            CreateSlider(fontSize, false);

            //Allow hiding the roles under the tracker name
            bShowRole = new JSONStorableBool("Show label with tracker #", false,
                (bool bRoleShow) =>
                {
                    trackEm.ShowRole(bRoleShow);
                    if (bRoleShow)
                    {
                        toggleShowLabel.toggle.isOn = false;
                        trackEm.ShowLabel(false);
                    }
                });
            RegisterBool(bShowRole);
            toggleShowRole=CreateToggle(bShowRole);
            
            //Show role as label
            bShowLabel = new JSONStorableBool("Show label only", false,
                (bool bLabelShow) => 
                {
                    trackEm.ShowLabel(bLabelShow);
                    if (bLabelShow)
                    {
                        toggleShowRole.toggle.isOn = false;
                        trackEm.ShowRole(false);
                    }
                });
            RegisterBool(bShowLabel);
            toggleShowLabel=CreateToggle(bShowLabel);

            //Create input fields
            foreach(JSONStorableString trackerName in trackEm.ToStorableString())
            {
                trackerLabels.Add(CreateInputField(trackerName, true));
            }
        }

        //Create labeled input field
        public UIDynamicTextField CreateInputField(JSONStorableString labeledInput, bool rhs = false, float fontScale=1.15f)
        {
            //Create top label
            UIDynamicTextField labelField = CreateTextField(new JSONStorableString("Label", $"<b>{labeledInput.name}:[{labeledInput.altName}]</b>"), rhs);
            labelField.backgroundColor = Color.clear;
            SetLayoutHeight(labelField, labelField.UItext.fontSize * fontScale);

            //Create a text field
            UIDynamicTextField textField = CreateTextField(labeledInput, rhs);
            textField.backgroundColor = Color.white;
            SetLayoutHeight(textField, textField.UItext.fontSize * fontScale);
            
            //Add an input field to text field
            InputField inField = textField.gameObject.AddComponent<InputField>();
            inField.textComponent = textField.UItext;
            labeledInput.inputField = inField;

            //Handle changes to input so it updates labels in the storable
            inField.onValueChanged.AddListener(
                (string labelChanged) => 
                {
                    labeledInput.val = labelChanged;
                    trackEm.UpdateText();
                });
            
            RegisterString(labeledInput);

            return textField;
        }
        
        //Set the height of the ui item
        public void SetLayoutHeight(UIDynamic uiItem, float height)
        {
            LayoutElement lay = uiItem.GetComponent<LayoutElement>();
            if (lay != null)
            {
                lay.minHeight = height;
                lay.preferredHeight = height;
            }
            uiItem.height = height;
        }
        #endregion SetupUI

        #region Main

        public TrackTrackers trackEm;

        public override void Init()
        {
            try
            {
                //Check if assigned as session plugin
                if (containingAtom.type!="CoreControl" && containingAtom.type!="SessionPluginManager")
                {
                    SuperController.LogError($"{pluginName} This must be used as a scene plugin not on an actor. Use Main Menu->Scene Plugin->Add Plugin.");
                    DestroyImmediate(this);
                    return;
                }
                
                //Check if VR is present
                if (!XRSettings.isDeviceActive)
                {
                    SuperController.LogError($"{pluginName} requires VR and trackers.");
                    DestroyImmediate(this);
                    return;
                }

                //Setup Plugin Name
                pluginLabelJSON.val = $"{pluginName} version={pluginVersion}";
                
                trackEm = new TrackTrackers();               
                SetupUI();
                trackEm.SetFontSize(fontSize.val);
            }
            catch(Exception e)
            {
                SuperController.LogError($"{pluginName}:Init:{e}");
            }
        }

        public void Update()
        {
            try
            {
                trackEm.DoUpdate();
            }
            catch(Exception e)
            {
                SuperController.LogError($"{pluginName}:Update:{e}");
            }
        }

        void OnDestroy()
        {
            try
            {
                trackEm?.DoDestroy();
            }
            catch(Exception e)
            {
                SuperController.LogError($"{pluginName}:OnDestroy:{e}");
            }
        }
        #endregion Main
    }
}
